# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from enum import IntEnum
from functools import wraps
from typing import Optional

import celpy
from pydantic import BaseModel

# Define the CEL environment
ENV = celpy.Environment(
    annotations={
        "msg": celpy.celtypes.MapType,
        "scope": celpy.celtypes.MapType,
        "topic": celpy.celtypes.StringType,
        "ts": celpy.celtypes.DoubleType,
    }
)


# Define the enums/constants related to validation errors
class ErrorSectionEnum(IntEnum):
    """Define the enumeration for the error section"""

    CONDITION = 1
    ACTION = 2


class ValidationErrorLocation(BaseModel):
    """Define the structure of the validation error location"""

    ruleIndex: int
    section: ErrorSectionEnum
    itemIndex: Optional[int] = None


class ValidationErrorUnexpectedVersion(BaseModel):
    """Define the structure of the validation error unexpected version"""

    allowedVersions: list[str]


class ValidationError(BaseModel):
    """Define the structure of the validation error"""

    location: Optional[ValidationErrorLocation] = None
    unexpectedVersion: Optional[ValidationErrorUnexpectedVersion] = None
    syntaxError: Optional[dict] = None
    emptySection: Optional[dict] = None


class ValidationResult(BaseModel):
    """Define the structure of the validation result"""

    success: bool
    errors: list[ValidationError]


def log_level_decorator(level):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Save the original logging level
            original_level = logging.getLogger().level
            # Set logging level to the specified level
            logging.getLogger().setLevel(level)
            try:
                # Execute the function
                return func(*args, **kwargs)
            finally:
                # Reset logging level to original
                logging.getLogger().setLevel(original_level)

        return wrapper

    return decorator



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

from ruleengine.dsl.action import Action
from ruleengine.dsl.base_conditions import concat, condition_start_time, ts
from ruleengine.dsl.condition import Condition


def noop_upload(
    trigger_ts: int,
    before: int,
    after: int,
    title: str,
    description: str,
    labels: List[str],
    extra_files: List[str],
    white_list: List[str],
):
    pass


def noop_create_moment(
    title: str,
    description: str,
    timestamp: int,
    start_time: int,
    create_task: bool,
    sync_task: bool,
    assign_to: Optional[str],
    custom_fields: Optional[str],
):
    pass


noop = {"upload": noop_upload, "create_moment": noop_create_moment}


class ForwardingAction(Action):
    def __init__(self, thunk, args):
        self.__thunk = thunk
        self.__args = args

    def run(self, item, scope):
        actual_args = {}
        for name, value in self.__args.items():
            if isinstance(value, Condition):
                new_value, _ = value.evaluate_condition_at(item, scope)
                actual_args[name] = new_value
            else:
                actual_args[name] = value
        self.__thunk(**actual_args)


def upload_factory(impl):
    def res(
        title=concat("Device auto upload @ ", ts),
        description="",
        labels=[],
        extra_files=[],
        white_list=[],
        before=10,
        after=0,
    ):
        args = {
            "before": before,
            "after": after,
            "title": Condition.map(Condition.wrap(title), str),
            "description": Condition.map(Condition.wrap(description), str),
            "labels": labels,
            "extra_files": extra_files,
            "white_list": white_list,
            "trigger_ts": ts,
        }

        return ForwardingAction(impl, args)

    return res


def create_moment_factory(impl):
    def res(
        title,
        description="",
        timestamp=ts,
        start_time=condition_start_time,
        create_task=False,
        sync_task=False,
        assign_to=None,
        custom_fields="",
    ):
        args = {
            "title": Condition.map(Condition.wrap(title), str),
            "description": Condition.map(Condition.wrap(description), str),
            "timestamp": Condition.map(Condition.wrap(timestamp), float),
            "start_time": Condition.map(Condition.wrap(start_time), float),
            "create_task": Condition.wrap(create_task),
            "sync_task": Condition.wrap(sync_task),
            "assign_to": assign_to,
            "custom_fields": Condition.map(Condition.wrap(custom_fields), str),
        }
        return ForwardingAction(impl, args)

    return res



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

from .base_conditions import and_, msg, msgtype, or_
from .condition import Condition

# TODO: Add tests
# TODO: Add other fields that are common for logs

_is_foxglove = or_(msgtype == "foxglove_msgs/Log", msgtype == "foxglove.Log")
_is_ros = msgtype == "rosgraph_msgs/Log"

LogLevel = Enum("LogLevel", ["UNKNOWN", "DEBUG", "INFO", "WARN", "ERROR", "FATAL"])

log = or_(
    and_(_is_ros, msg.msg),
    and_(_is_foxglove, msg.message),
    # Default case
    "",
)

log_level = or_(
    and_(_is_ros, Condition.map(msg, lambda m: ros_log_level(m.level))),
    and_(_is_foxglove, Condition.map(msg, lambda m: foxglove_log_level(m.level))),
    # Default case
    LogLevel.UNKNOWN,
)


def ros_log_level(num):
    if num == 1:
        return LogLevel.DEBUG
    elif num == 2:
        return LogLevel.INFO
    elif num == 4:
        return LogLevel.WARN
    elif num == 8:
        return LogLevel.ERR
    elif num == 16:
        return LogLevel.FATAL
    else:
        return LogLevel.UNKNOWN


def foxglove_log_level(num):
    if num == 1:
        return LogLevel.DEBUG
    elif num == 2:
        return LogLevel.INFO
    elif num == 3:
        return LogLevel.WARN
    elif num == 4:
        return LogLevel.ERROR
    elif num == 5:
        return LogLevel.FATAL
    else:
        return LogLevel.UNKNOWN


__all__ = [
    "log",
    "log_level",
    "LogLevel",
]



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base_conditions import and_
from .condition import Condition


def sustained(context_condition, variable_condition, duration):
    """
    This condition triggers when the variable condition continues to be true for
    the given duration.

    The context condition is used to limit the scope of the variable condition.
    For example, we might say "if topic X has value Y for 10 seconds", which
    translates to context being topic == X, and variable being value == Y
    """
    return and_(
        context_condition,
        RisingEdgeCondition(SustainedCondition(variable_condition, duration)),
    )


def sequential(*condition_factories, duration=None):
    return SequenceMatchCondition(list(condition_factories), duration)


def timeout(*condition_factories, duration):
    return SequenceMatchCondition(list(condition_factories), duration, True)


def any_order(*condition_factories, reset_time=None):
    return RisingEdgeCondition(AnyOrderCondition(list(condition_factories), reset_time))


def repeated(condition_factory, /, times, duration):
    assert times > 1, "In repeated condition, times must be more than 1"
    return and_(
        condition_factory(),
        RisingEdgeCondition(RepeatedCondition(condition_factory(), times, duration)),
    )


def debounce(condition_factory, /, duration):
    return and_(condition_factory(), RisingEdgeCondition(condition_factory(), duration))


def throttle(condition, duration):
    return ThrottleCondition(condition, duration)


class RisingEdgeCondition(Condition):
    """
    A condition that detects when child condition changes from false to true.

    Once child condition becomes true, this condition will not fire unless the
    child condition first becomes false, then back to true.

    If max_gap is given, and two invocations of this condition have timestamps
    further apart than the gap, the state is reset. It behaves as if a false is
    inserted in the gap.
    """

    def __init__(self, condition, max_gap=None):
        assert isinstance(condition, Condition)
        self.__condition = condition
        self.__active = False
        self.__last_activation = None
        self.__max_gap = max_gap

    def evaluate_condition_at(self, item, scope):
        value, new_scope = self.__condition.evaluate_condition_at(item, scope)
        if not value:
            self.__active = False
            return False, new_scope

        if self.__active and (
            self.__max_gap is None or item.ts - self.__last_activation < self.__max_gap
        ):
            self.__last_activation = item.ts
            return False, new_scope

        self.__active = True
        self.__last_activation = item.ts
        return True, new_scope


class SustainedCondition(Condition):
    """
    This condition triggers when the child condition is true for the given
    duration.
    The state is reset once the child condition becomes false.
    """

    def __init__(self, condition, duration=-1):
        super().__init__()

        assert isinstance(condition, Condition)
        self.__condition = condition
        self.__duration = duration
        self.__start = None

    def evaluate_condition_at(self, item, scope):
        value, new_scope = self.__condition.evaluate_condition_at(item, scope)
        if not value:
            self.__start = None
            return False, new_scope

        if self.__start is None:
            self.__start = item.ts

        if item.ts - self.__start > self.__duration:
            return True, {**new_scope, "start_time": self.__start}

        return False, new_scope


class SequenceMatchCondition(Condition):
    """
    This condition triggers when the child conditions are true in the given
    order, and within the given duration if set.

    If trigger_on_timeout is set, the condition will only trigger when the
    sequence is not matched within the given duration.

    Note that the sequence input is a list of condition factories, not conditions.

    Also note that we do not support overlapping sequences.
    """

    def __init__(self, factory_sequence, duration=None, trigger_on_timeout=False):
        super().__init__()

        assert len(factory_sequence) > 1, "Sequence must be longer than 1"
        for factory in factory_sequence:
            assert isinstance(factory(), Condition)

        self.__factory_seq = factory_sequence
        self.__duration = duration
        self.__trigger_on_timeout = trigger_on_timeout
        self.__current_scope = None
        self.__start_time = None
        self._reset()

    def _reset(self):
        self.__start_time = None
        self.__current_index = 0
        self.__current_scope = None
        self.__seq = [factory() for factory in self.__factory_seq]

    def evaluate_condition_at(self, item, scope):
        if (
            self.__duration is not None
            and self.__start_time is not None
            and item.ts - self.__start_time > self.__duration
        ):
            ret = bool(self.__trigger_on_timeout), {
                **self.__current_scope,
                "start_time": self.__start_time,
            }
            self._reset()
            return ret

        matched, new_scope = self.__seq[self.__current_index].evaluate_condition_at(
            item, self.__current_scope or scope
        )
        if matched:
            self.__current_index += 1
            if self.__current_index == len(self.__seq):
                ret = not self.__trigger_on_timeout, {
                    **new_scope,
                    "start_time": self.__start_time,
                }
                self._reset()
                return ret

            self.__current_scope = new_scope
            if self.__start_time is None:
                self.__start_time = item.ts

        return False, scope


class RepeatedCondition(Condition):
    """
    This condition triggers when the child condition is true for the given
    number of times within the given duration.
    """

    def __init__(self, condition, times, duration):
        assert isinstance(condition, Condition)
        self.__condition = condition
        self.__times = times
        self.__duration = duration
        self.__trigger_times = []

    def evaluate_condition_at(self, item, scope):
        value, new_scope = self.__condition.evaluate_condition_at(item, scope)

        if not value:
            return False, scope

        self.__trigger_times.append(item.ts)
        self.__trigger_times = [
            t for t in self.__trigger_times if item.ts - t <= self.__duration
        ]

        if len(self.__trigger_times) >= self.__times:
            return True, {**new_scope, "start_time": self.__trigger_times[0]}

        return False, scope


class ThrottleCondition(Condition):
    """
    This condition triggers when the child condition is true, but only if the
    last trigger is more than the given duration ago.
    """

    def __init__(self, condition, duration):
        assert isinstance(condition, Condition)
        self.__condition = condition
        self.__duration = duration
        self.__last_trigger = -duration

    def evaluate_condition_at(self, item, scope):
        if item.ts - self.__last_trigger < self.__duration:
            return False, scope

        value, scope = self.__condition.evaluate_condition_at(item, scope)
        if value:
            self.__last_trigger = item.ts
            return value, scope

        return False, scope


class AnyOrderCondition(Condition):
    """
    This condition triggers when all the child conditions are true, but the
    order of the conditions does not matter.
    The condition will reset once triggered or when the reset_time has passed after the first condition is met.
    """

    def __init__(self, condition_factories, reset_time=None):
        self.__factories = condition_factories
        self.__reset_time = reset_time
        self.__unsatisfied = [factory() for factory in condition_factories]
        self.__curr_scope = None
        self.__start_time = None

    def evaluate_condition_at(self, item, scope):
        if (
            self.__reset_time is not None
            and self.__start_time is not None
            and item.ts - self.__start_time > self.__reset_time
        ):
            self.reset()

        if self.__curr_scope is None:
            self.__curr_scope = scope

        new_conditions = []
        for c in self.__unsatisfied:
            value, new_scope = c.evaluate_condition_at(item, self.__curr_scope)
            if value:
                self.__curr_scope = new_scope
                if self.__start_time is None:
                    self.__start_time = item.ts
            else:
                new_conditions.append(c)

        if not new_conditions:
            self.reset()
            return True, self.__curr_scope

        self.__unsatisfied = new_conditions
        return False, self.__curr_scope

    def reset(self):
        self.__unsatisfied = [factory() for factory in self.__factories]
        self.__start_time = None


__all__ = [
    "any_order",
    "debounce",
    "repeated",
    "sequential",
    "sustained",
    "throttle",
    "timeout",
]



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from collections import namedtuple

from ruleengine.dsl.action import Action
from ruleengine.engine import DiagnosisItem, Engine, Rule
from tests.dsl.utils import str_to_condition

MockMessage = namedtuple(
    "MockMessage", "int_value str_value list_value list_object_value"
)
ListMessage = namedtuple("ListMessage", "value")
simple_sequence = [
    DiagnosisItem(
        "t1",
        MockMessage(1, "hello", [0, 1], []),
        0,
        "MockMessage",
    ),
    DiagnosisItem(
        "t2",
        MockMessage(
            2, "hello", [0, 1, 2], [ListMessage(0), ListMessage(1), ListMessage(2)]
        ),
        1,
        "MockMessage",
    ),
    DiagnosisItem(
        "t1",
        MockMessage(
            3,
            "heLlo",
            [0, 1, 2, 3],
            [ListMessage(0), ListMessage(1), ListMessage(2), {"value": 3}],
        ),
        2,
        "MockMessage",
    ),
    DiagnosisItem(
        "t2",
        MockMessage(
            4,
            "hello",
            [0, 1, 2, 3, 4],
            [
                ListMessage(0),
                ListMessage(1),
                ListMessage(2),
                ListMessage(3),
                ListMessage(4),
            ],
        ),
        3,
        "MockMessage",
    ),
    DiagnosisItem(
        "t2",
        MockMessage(
            5,
            "world",
            [0, 1, 2, 3, 4, 5],
            [
                ListMessage(0),
                ListMessage(1),
                ListMessage(2),
                ListMessage(3),
                ListMessage(4),
                ListMessage(5),
            ],
        ),
        4,
        "MockMessage",
    ),
    DiagnosisItem(
        "t3",
        MockMessage(
            5, "The value is 324, which is expected to be less than 22", [], []
        ),
        5,
        "MockMessage",
    ),
    DiagnosisItem(
        "t3",
        MockMessage(5, "The value is 11, which is expected to be less than 22", [], []),
        5,
        "MockMessage",
    ),
]


class CollectAction(Action):
    def __init__(self):
        self.collector = []

    def run(self, item, scope):
        self.collector.append(item)


class BaseConditionTest(unittest.TestCase):
    def test_always(self):
        result = self.__run_test("always")
        self.assertEqual(len(result), 7, result)

    def test_msg(self):
        result = self.__run_test('msg == MockMessage(1, "hello", [0, 1], [])')
        self.assertEqual(len(result), 1, result)

    def test_ts(self):
        result = self.__run_test("ts == 0")
        self.assertEqual(len(result), 1, result)

    def test_topic(self):
        result = self.__run_test('topic == "t1"')
        self.assertEqual(len(result), 2, result)

    def test_msgtype(self):
        result = self.__run_test('msgtype == "MockMessage"')
        self.assertEqual(len(result), 7, result)

        result = self.__run_test('msgtype == "FalseMessage"')
        self.assertEqual(len(result), 0, result)

    def test_none_values(self):
        result = self.__run_test('msg.this.doesnt.exist == "MockMessage"')
        self.assertEqual(len(result), 0, result)

        result = self.__run_test("is_none(msg.this.doesnt.exist)")
        self.assertEqual(len(result), 7, result)

    def test_list_values(self):
        result = self.__run_test("msg.list_value[2] == 2")
        self.assertEqual(len(result), 4, result)

        result = self.__run_test("3 in map_attr(msg.list_object_value, 'value')")
        self.assertEqual(len(result), 3, result)

    def test_and(self):
        result = self.__run_test('msgtype == "MockMessage" and topic == "t1"')
        self.assertEqual(len(result), 2, result)

    def test_or(self):
        result = self.__run_test('ts == 1 or topic == "t1"')
        self.assertEqual(len(result), 3, result)

    def test_not(self):
        result = self.__run_test("not ts == 1")
        self.assertEqual(len(result), 6, result)

    def test_topic_match(self):
        result = self.__run_test('topic == "t1"')
        self.assertEqual(len(result), 2, result)

    def test_type_match(self):
        result = self.__run_test('msgtype == "MockMessage"')
        self.assertEqual(len(result), 7, result)

        result = self.__run_test('msgtype == "FalseMessage"')
        self.assertEqual(len(result), 0, result)

    def test_complex_conditions(self):
        result = self.__run_test('topic == "t2" and msg.int_value > 2')
        self.assertEqual(len(result), 2, result)

    def test_function_calls(self):
        result = self.__run_test('msg.str_value.upper() == "HELLO"')
        self.assertEqual(len(result), 4, result)

    def test_get_set_values(self):
        result = self.__run_test(
            'set_value("somekey", msg.str_value) and get_value("somekey") == "hello"'
        )
        self.assertEqual(len(result), 3, result)

        result = self.__run_test('get_value("non-exist-key") == 1')
        self.assertEqual(len(result), 0, result)

    def test_has(self):
        result = self.__run_test(
            '"el" in msg.str_value and get_value("cos/contains") == "el"'
        )
        self.assertEqual(len(result), 3, result)
        result = self.__run_test(
            '"el" in msg.str_value and get_value("cos/contains") == "ee"'
        )
        self.assertEqual(len(result), 0, result)

    def test_regex(self):
        result = self.__run_test(
            'regex(msg.str_value, r"e[lL]lo").group(0) == "ello"',
        )
        self.assertEqual(len(result), 3, result)
        result = self.__run_test(
            'regex(msg.str_value, r"e[lL]lo").group(0) == "eLlo"',
        )
        self.assertEqual(len(result), 1, result)

    def test_concat(self):
        result = self.__run_test(
            'concat(msg.str_value, "---", msg.int_value) == "hello---1"',
        )
        self.assertEqual(len(result), 1, result)

    def test_coerce(self):
        result = self.__run_test(
            """regex(
                msg.str_value,
                r"The value is (\\d+), which is expected to be less than",
            ).group(1) > 111""",
        )
        self.assertEqual(len(result), 1, result)

    @staticmethod
    def __run_test(expr_str):
        action = CollectAction()
        engine = Engine(
            [
                Rule(
                    [str_to_condition(expr_str, {"MockMessage": MockMessage})],
                    [action],
                    {},
                )
            ]
        )
        for item in simple_sequence:
            engine.consume_next(item)
        return action.collector



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from ruleengine.dsl.condition import Condition


class ConditionTest(unittest.TestCase):
    def test_arithmetics(self):
        self.assertEqual(-4, int(self._get_eval_result(-Condition.wrap(4))))
        self.assertEqual(4, int(self._get_eval_result(+Condition.wrap(4))))

        self.assertEqual(10, int(self._get_eval_result(Condition.wrap(4) + 6)))
        self.assertEqual(10, int(self._get_eval_result(6 + Condition.wrap(4))))

        self.assertEqual(-2, int(self._get_eval_result(Condition.wrap(4) - 6)))
        self.assertEqual(2, int(self._get_eval_result(6 - Condition.wrap(4))))

        self.assertEqual(24, int(self._get_eval_result(Condition.wrap(4) * 6)))
        self.assertEqual(24, int(self._get_eval_result(6 * Condition.wrap(4))))

        self.assertEqual(2, int(self._get_eval_result(Condition.wrap(12) / 6)))
        self.assertEqual(2, int(self._get_eval_result(12 / Condition.wrap(6))))

    def _get_eval_result(self, cond):
        return cond.evaluate_condition_at(None, {})[0]



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import celpy.adapter

from rule_engine.action import Action
from rule_engine.condition import Condition
from rule_engine.utils import (
    ErrorSectionEnum,
    ValidationError,
    ValidationErrorLocation,
    ValidationResult,
)

ALLOWED_VERSIONS = ["v2"]


class Rule:
    """
    Defining a rule with a condition and corresponding actions

    RuleSpec should be of the following structure:
    {
        "version": "v2",
        "rules": [
            {
                "conditions": [
                    "msg['temperature'] > 20",
                    "int(msg['humidity']) < int(30)"
                ],
                "actions: [
                    {
                        "name": "action_1",
                        "kwargs": {
                            "arg1_1": "{msg['item']}",
                            "arg1_2": 1
                        }
                    },
                    {
                        "name": "action_2",
                        "kwargs": {
                            "arg2_1": "{msg.code}",
                            "arg2_2": 1
                        }
                    }
                ],
                "scopes": [
                    {
                        "scope_key_a": "scope_value_a_1"
                        "scope_key_b": "scope_value_b_1"
                    },
                    {
                        "scope_key_a": "scope_value_a_2"
                        "scope_key_b": "scope_value_b_2"
                    }
                ],
                "topics": ["topic_1", "topic_2"]
            }
        ]
    }
    Note that for each scope item will be used to construct a separate rule,
    In the above example, two rules will be created with the same conditions and actions and topics,
    but with one scope being {"scope_key_a": "scope_value_a_1", "scope_key_b": "scope_value_b_1"}
    and the other being {"scope_key_a": "scope_value_a_2", "scope_key_b": "scope_value_b_2"}
    Also be aware that the rule_idx of scope created rules will be the same as the original rule
    """

    def __init__(
        self,
        raw: any,
        conditions: list[Condition],
        actions: list[Action],
        scope: dict[str, str],
        topics: list[str],
        metadata: dict[str, any] = None,  # user-defined metadata
    ):
        self.raw = raw
        self.conditions = conditions
        self.actions = actions
        self.scope = celpy.adapter.json_to_cel(scope)
        self.topics = topics
        self.metadata = metadata or {}


def validate_rules_spec(
    rules_spec: dict[str, any], action_impls: dict[str, any]
) -> tuple[list[Rule], ValidationResult]:
    """
    Validate rules spec, this function will only validate the structure of the rules spec
    """
    all_rules = []
    errors = []
    for rule_idx, rule_spec in enumerate(rules_spec.get("rules", [])):
        rules, errs = validate_rule_spec(rule_spec, action_impls, rule_idx)
        all_rules.extend(rules)
        errors.extend(errs)

    return all_rules, ValidationResult(success=not errors, errors=errors)


def validate_rule_spec(
    rule_spec: dict[str, any], action_impls: dict[str, any], rule_idx: int
) -> tuple[list[Rule], list[ValidationError]]:
    """
    Validate single rule spec
    """
    errors: list[ValidationError] = []

    if not rule_spec.get("conditions", []):
        errors.append(
            ValidationError(
                location=ValidationErrorLocation(
                    ruleIndex=rule_idx, section=ErrorSectionEnum.CONDITION
                ),
                emptySection={},
            ),
        )

    conditions = []
    for condition_idx, condition_spec in enumerate(rule_spec.get("conditions", [])):
        condition, err = Condition.compile_and_validate(condition_spec)
        if err:
            errors.append(
                ValidationError(
                    location=ValidationErrorLocation(
                        ruleIndex=rule_idx,
                        section=ErrorSectionEnum.CONDITION,
                        itemIndex=condition_idx,
                    ),
                    syntaxError={},
                ),
            )
        else:
            conditions.append(condition)

    if not rule_spec.get("actions", []):
        errors.append(
            ValidationError(
                location=ValidationErrorLocation(
                    ruleIndex=rule_idx, section=ErrorSectionEnum.ACTION
                ),
                emptySection={},
            ),
        )

    actions = []
    for action_idx, action_spec in enumerate(rule_spec.get("actions", [])):
        action_name = action_spec.get("name", "")
        action_kwargs = action_spec.get("kwargs", {})
        action_impl = action_impls.get(action_name)
        action, err = Action.compile_and_validate(
            action_name, action_kwargs, action_impl
        )
        if err:
            errors.append(
                ValidationError(
                    location=ValidationErrorLocation(
                        ruleIndex=rule_idx,
                        section=ErrorSectionEnum.ACTION,
                        itemIndex=action_idx,
                    ),
                    syntaxError={},
                ),
            )
        else:
            actions.append(action)

    if errors:
        return [], errors

    scopes = rule_spec.get("scopes", [])
    topics = rule_spec.get("topics", [])
    if not scopes:
        scopes = [{}]

    rules = []
    for scope in scopes:
        rules.append(Rule(rule_spec, conditions, actions, scope, topics))

    return rules, []



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from ruleengine.dsl.validation.validation_result import ValidationErrorType
from ruleengine.dsl.validation.validator import validate_action, validate_condition
from ruleengine.engine import Rule

ALLOWED_VERSIONS = ["v1"]


def validate_config(config, action_impls, project_name=""):
    """
    Validate a rule specification. Use validate_config_wrapped instead if action_impls depends on the rule.

    It is assumed that the input config is a DiagnosisRule specified in

    https://github.com/coscene-io/cosceneapis/blob/main/coscene/dataplatform/v1alpha2/common/diagnosis_rule.proto

    and this function will return a DiagnosisRuleValidationResult specified in

    https://github.com/coscene-io/cosceneapis/blob/main/coscene/dataplatform/v1alpha2/common/diagnosis_rule_validation_result.proto

    :param config: The rule specification.
    :param action_impls: A dictionary of action implementations.
    :param project_name: The name of the project that the rule is associated with.
    """
    # action_impls_wrapped = {k: lambda _: v for k, v in action_impls.items()}
    action_impls_wrapped = {}
    for k, v in action_impls.items():
        action_impls_wrapped[k] = lambda _: v
    return validate_config_wrapped(config, action_impls_wrapped, project_name)


def validate_config_wrapped(config, action_impls_wrapped, project_name=""):
    """
    Validate a rule specification where the action implementations depend on the rule.
    """

    # TODO: Instead of putting together these objects by hand, we should connect
    # the protos generated in cosceneapis

    raw_version = config.get("version", "")
    raw_rules = config.get("rules", [])

    if raw_version not in ALLOWED_VERSIONS:
        return {
            "success": False,
            "errors": [{"unexpectedVersion": {"allowedVersions": ALLOWED_VERSIONS}}],
        }, []

    errors = []
    rules = []
    for i, rule in enumerate(raw_rules):
        action_impls = {k: v(rule) for k, v in action_impls_wrapped.items()}
        rule_errors, new_rules = _validate_rule(rule, i, action_impls, project_name)
        errors += rule_errors
        rules += new_rules

    success = not bool(errors)
    return {"success": success, "errors": errors}, rules


def _validate_rule(rule, rule_index, action_impls, project_name):
    errors = []
    raw_conditions = rule.get("when", [])
    raw_actions = rule.get("actions", [])
    if not raw_conditions:
        errors.append(
            {
                "location": {
                    "ruleIndex": rule_index,
                    "section": 1,
                },
                "emptySection": {},
            }
        )
    if not raw_actions:
        errors.append(
            {
                "location": {
                    "ruleIndex": rule_index,
                    "section": 2,
                },
                "emptySection": {},
            }
        )

    def parse_rule():
        conditions = []
        for i, cond_str in enumerate(raw_conditions):
            res = validate_condition(cond_str)
            if not res.success:
                errors.append(
                    {
                        "location": {
                            "ruleIndex": rule_index,
                            "section": 1,
                            "itemIndex": i,
                            "details": res.details,
                        },
                        **_convert_to_json_error(res),
                    }
                )
            else:
                conditions.append(res.entity)

        actions = []
        for i, action_str in enumerate(raw_actions):
            res = validate_action(action_str, action_impls)
            if not res.success:
                errors.append(
                    {
                        "location": {
                            "ruleIndex": rule_index,
                            "section": 2,
                            "itemIndex": i,
                            "details": res.details,
                        },
                        **_convert_to_json_error(res),
                    }
                )
            else:
                actions.append(res.entity)
        return conditions, actions

    # We parse the rules once to see if there are any errors. If so, bail. Also,
    # if there are no `each` values, we use this set of parsed values for the
    # actual execution.
    #
    # If there are `each` values, we parse a new set of rules for each of them,
    # since rules are stateful, so we want separate instances.

    conditions, actions = parse_rule()
    upload_limit = rule.get("upload_limit", {})
    if errors:
        return errors, []

    templating_args = rule.get("each", [])
    if not templating_args:
        return [], [
            Rule(
                conditions, actions, {}, upload_limit, copy.deepcopy(rule), project_name
            )
        ]

    new_rules = []
    for arg in templating_args:
        conditions, actions = parse_rule()
        new_rules.append(
            Rule(
                conditions,
                actions,
                arg,
                upload_limit,
                {**copy.deepcopy(rule), "each": [arg]},
                project_name,
            )
        )
    return [], new_rules


def _convert_to_json_error(result):
    if result.error_type == ValidationErrorType.SYNTAX or ValidationErrorType.EMPTY:
        return {"syntax_error": {}}
    elif result.error_type == ValidationErrorType.NOT_CONDITION:
        return {"notCondition": {"actualType": result.details["actual"]}}
    elif result.error_type == ValidationErrorType.NOT_ACTION:
        return {"notAction": {"actualType": result.details["actual"]}}
    elif result.error_type == ValidationErrorType.UNDEFINED:
        return {"nameUndefined": {"name": result.details["name"]}}
    elif result.error_type == ValidationErrorType.UNKNOWN:
        return {"genericError": {"msg": result.details["message"]}}
    else:
        raise Exception(f"Unknown error type: {result.error_type}")



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from rule_engine.rule import validate_rules_spec
from rule_engine.utils import ErrorSectionEnum


def _serialize_impl(str_arg, int_arg):
    pass


class TestRule(unittest.TestCase):
    def test_compile_and_validate_success(self):
        rule_spec = {
            "rules": [
                {
                    "conditions": [
                        "msg['temperature'] > 20",
                        "msg['humidity'] > 20",
                    ],
                    "actions": [
                        {
                            "name": "serialize",
                            "kwargs": {
                                "str_arg": "{msg['item']}",
                                "int_arg": 1,
                            },
                        },
                        {
                            "name": "serialize",
                            "kwargs": {
                                "str_arg": "{msg['item']}",
                                "int_arg": 1,
                            },
                        },
                    ],
                    "scopes": [],
                    "topics": ["test"],
                }
            ]
        }

        rules, result = validate_rules_spec(rule_spec, {})
        self.assertDictEqual(
            result.model_dump(exclude_unset=True),
            {
                "success": True,
                "errors": [],
            },
        )

    def test_compile_and_validate_empty_condition(self):
        rule_spec = {
            "rules": [
                {
                    "conditions": [
                        "",
                    ],
                    "actions": [
                        {
                            "name": "serialize",
                            "kwargs": {
                                "str_arg": "{msg['item']}",
                                "int_arg": 1,
                            },
                        },
                    ],
                    "scopes": [],
                    "topics": ["test"],
                }
            ]
        }
        rules, result = validate_rules_spec(rule_spec, {})
        self.assertDictEqual(
            result.model_dump(exclude_unset=True),
            {
                "success": False,
                "errors": [
                    {
                        "location": {
                            "ruleIndex": 0,
                            "section": ErrorSectionEnum.CONDITION,
                            "itemIndex": 0,
                        },
                        "syntaxError": {},
                    }
                ],
            },
        )

    def test_compile_and_validate_empty_action(self):
        rule_spec = {
            "rules": [
                {
                    "conditions": [
                        "msg['temperature'] > 20",
                    ],
                    "actions": [],
                    "scopes": [],
                    "topics": ["test"],
                }
            ]
        }
        rules, result = validate_rules_spec(rule_spec, {})
        self.assertDictEqual(
            result.model_dump(exclude_unset=True),
            {
                "success": False,
                "errors": [
                    {
                        "location": {
                            "ruleIndex": 0,
                            "section": ErrorSectionEnum.ACTION,
                        },
                        "emptySection": {},
                    }
                ],
            },
        )

    def test_compile_and_validate_multiple_errors(self):
        rule_spec = {
            "rules": [
                {
                    "conditions": [
                        "msg['temperature'] > 20",
                        "msg['humidity'] > ",
                    ],
                    "actions": [
                        {
                            "name": "serialize",
                            "kwargs": {
                                "str_arg": "{msg['item']}",
                                "int_arg": 1,
                            },
                        },
                        {
                            "name": "serialize",
                            "kwargs": {
                                "str_arg": "{msg[}",
                                "int_arg": 1,
                            },
                        },
                        {
                            "name": "serialize",
                            "kwargs": {
                                "str_arg": "{msg['item']}",
                                "int_arg": 1,
                            },
                        },
                    ],
                    "scopes": [],
                    "topics": ["test"],
                }
            ]
        }
        rules, result = validate_rules_spec(rule_spec, {})
        self.assertDictEqual(
            result.model_dump(exclude_unset=True),
            {
                "success": False,
                "errors": [
                    {
                        "location": {
                            "ruleIndex": 0,
                            "section": ErrorSectionEnum.CONDITION,
                            "itemIndex": 1,
                        },
                        "syntaxError": {},
                    },
                    {
                        "location": {
                            "ruleIndex": 0,
                            "section": ErrorSectionEnum.ACTION,
                            "itemIndex": 1,
                        },
                        "syntaxError": {},
                    },
                ],
            },
        )



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import celpy

from rule_engine.rule import Rule


class Engine:
    """
    The rule engine represents a collection of rules
    """

    def __init__(self, rules: list[Rule]):
        self.rules = rules
        self.cur_activation = None

    def example_consume_next(self, msg: dict[str, any], topic: str, ts: float):
        """
        An example of how to consume a message and trigger rules
        In a real-world scenario, there might be more complex logic before condition evaluation
        before action running, or after action running. Use this as a reference.
        """
        self.load_message(msg, topic, ts)
        for rule_idx, _ in enumerate(self.rules):
            if self.evaluate_rule_condition(rule_idx):
                self.run_rule_actions(rule_idx)

    def load_message(self, msg: dict[str, any], topic: str, ts: float):
        """Load the message into the engine and prepared for rule evaluation"""
        self.cur_activation = {
            "msg": celpy.adapter.json_to_cel(msg),
            "topic": celpy.celtypes.StringType(topic),
            "ts": celpy.celtypes.DoubleType(ts),
        }

    def evaluate_rule_condition(self, rule_idx):
        """Evaluate the condition of a rule against the current activation"""
        rule = self.rules[rule_idx]
        activation = {
            **self.cur_activation,
            "scope": rule.scope,
        }
        return all(cond.evaluate(activation) for cond in rule.conditions)

    def run_rule_actions(self, rule_idx):
        """Run the actions of a rule against the current activation"""
        rule = self.rules[rule_idx]
        activation = {
            **self.cur_activation,
            "scope": rule.scope,
        }
        for action in rule.actions:
            action.run(activation)



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from ruleengine.dsl.action import Action
from ruleengine.dsl.condition import Condition

ValidationErrorType = Enum(
    "ValidationErrorType",
    ["SYNTAX", "EMPTY", "NOT_CONDITION", "NOT_ACTION", "UNDEFINED", "UNKNOWN"],
)


@dataclass
class ValidationResult:
    success: bool

    # If not success, fill in error type and details
    error_type: Optional[ValidationErrorType] = None
    details: dict = field(default_factory=dict)

    # If success, fill in validated entity
    entity: Optional[Action or Condition] = None

    def __post_init__(self):
        if self.success:
            assert self.entity is not None
        else:
            assert self.error_type



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cProfile
from sys import argv

from ruleengine.dsl.validation.config_validator import validate_config
from ruleengine.engine import DiagnosisItem, Engine

SAMPLE_CONFIG = {
    "version": "v1",
    "rules": [
        {
            "when": [
                '"error 1" in msg',
                '"error 21444444" in msg or "asioehtnai" in msg or "aioentaihe" in msg',
                '"error 21444444" in msg or "asioehtnai" in msg or "aioentaihe" in msg',
                '"error 21444444" in msg or "asioehtnai" in msg or "aioentaihe" in msg',
                '"error 21444444" in msg or "asioehtnai" in msg or "aioentaihe" in msg',
            ],
            "actions": ["upload()"],
        }
    ],
}


def sample_action(*args, **kwargs):
    print(args, kwargs)


if __name__ == "__main__":
    res, rules = validate_config(
        SAMPLE_CONFIG, {"upload": sample_action, "create_moment": sample_action}
    )
    assert res["success"], res
    engine = Engine(rules)

    with open(argv[1]) as f:
        lines = [
            DiagnosisItem(topic=argv[1], msg=line, ts=0, msgtype="log") for line in f
        ]

    def go():
        for item in lines:
            engine.consume_next(item)

    cProfile.run("go()")



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from sys import argv

from rule_engine.rule import (
    ALLOWED_VERSIONS as ALLOWED_VERSIONS_2,
    validate_rule_spec,
)
from rule_engine.utils import (
    ValidationError,
    ValidationErrorUnexpectedVersion,
    ValidationResult,
)
from ruleengine.dsl.base_actions import noop
from ruleengine.dsl.validation.config_validator import (
    ALLOWED_VERSIONS as ALLOWED_VERSIONS_1,
    validate_config,
)

ALLOWED_VERSIONS = ALLOWED_VERSIONS_1 + ALLOWED_VERSIONS_2


def main(rule_set_spec_str: str):
    rule_set_spec = json.loads(rule_set_spec_str)

    if not rule_set_spec.get("version", "") in ALLOWED_VERSIONS:
        print(
            ValidationResult(
                success=True,
                errors=[
                    ValidationError(
                        unexpectedVersion=ValidationErrorUnexpectedVersion(
                            allowedVersions=ALLOWED_VERSIONS,
                        )
                    )
                ],
            ).model_dump_json(exclude_unset=True)
        )
        exit(1)

    if rule_set_spec.get("version") in ALLOWED_VERSIONS_1:
        result, _ = validate_config(rule_set_spec, noop)
        print(json.dumps(result))
        exit(0 if result["success"] else 1)

    if rule_set_spec.get("version") in ALLOWED_VERSIONS_2:
        errors = []
        for rule_idx, rule_spec in enumerate(rule_set_spec.get("rules", [])):
            _, errs = validate_rule_spec(rule_spec, {}, rule_idx)
            errors.extend(errs)
        result = ValidationResult(success=not errors, errors=errors)
        print(result.model_dump_json(exclude_unset=True))
        exit(0 if result.success else 1)


if __name__ == "__main__":
    main(argv[1])



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from functools import partial

from rule_engine.engine import Engine
from rule_engine.rule import validate_rules_spec


class TestEngine(unittest.TestCase):
    @staticmethod
    def serialize_impl(str_arg, int_arg, result_buffer: list[dict]):
        result = {
            "str_arg": str_arg,
            "int_arg": int_arg,
        }
        result_buffer[0] = result

    @staticmethod
    def build_serialize_engine_from_spec(spec, buffer):
        rules, _ = validate_rules_spec(
            spec,
            {"serialize": partial(TestEngine.serialize_impl, result_buffer=buffer)},
        )
        return Engine(rules)

    def test_engine_comprehensive_single_condition(self):
        spec = {
            "rules": [
                {
                    "conditions": [
                        "msg.code > 20",
                    ],
                    "actions": [
                        {
                            "name": "serialize",
                            "kwargs": {
                                "str_arg": "{msg.code}",
                                "int_arg": 1,
                            },
                        },
                    ],
                    "scopes": [],
                    "topics": ["test_topic"],
                }
            ]
        }
        result_buffer = [{}]
        engine = self.build_serialize_engine_from_spec(spec, result_buffer)

        engine.example_consume_next({"code": 20}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {},
        )
        engine.example_consume_next({"code": 21}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {
                "str_arg": "21",
                "int_arg": 1,
            },
        )
        engine.example_consume_next({"code": 22}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {
                "str_arg": "22",
                "int_arg": 1,
            },
        )
        engine.example_consume_next({"code": 20}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {
                "str_arg": "22",
                "int_arg": 1,
            },
        )

    def test_engine_comprehensive_multiple_conditions(self):
        spec = {
            "rules": [
                {
                    "conditions": ["msg.code > 20", "int(msg.level) <= int(3)"],
                    "actions": [
                        {
                            "name": "serialize",
                            "kwargs": {
                                "str_arg": "{msg.code}",
                                "int_arg": 1,
                            },
                        },
                    ],
                    "scopes": [{}],
                    "topics": ["test_topic"],
                }
            ]
        }
        result_buffer = [{}]
        engine = self.build_serialize_engine_from_spec(spec, result_buffer)

        engine.example_consume_next({"code": 21, "level": 4}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {},
        )
        engine.example_consume_next({"code": 19, "level": 2}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {},
        )
        engine.example_consume_next({"code": 19, "level": 4}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {},
        )
        engine.example_consume_next({"code": 21, "level": 2}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {
                "str_arg": "21",
                "int_arg": 1,
            },
        )
        engine.example_consume_next({"code": 23, "level": 1}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {
                "str_arg": "23",
                "int_arg": 1,
            },
        )
        engine.example_consume_next({"code": 21, "level": 4}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {
                "str_arg": "23",
                "int_arg": 1,
            },
        )
        engine.example_consume_next({"code": 19, "level": 2}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {
                "str_arg": "23",
                "int_arg": 1,
            },
        )
        engine.example_consume_next({"code": 19, "level": 4}, "test_topic", 0.0)
        self.assertDictEqual(
            result_buffer[0],
            {
                "str_arg": "23",
                "int_arg": 1,
            },
        )



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

import celpy

from rule_engine.condition import Condition

TestActivation = {
    "msg": celpy.adapter.json_to_cel(
        {"message": {"code": 200}, "lst": [{"code": 1}, {"code": 2}, {"code": 3}]}
    ),
    "scope": celpy.adapter.json_to_cel({"code": 200}),
    "topic": celpy.celtypes.StringType("/TestTopic"),
    "ts": celpy.celtypes.DoubleType(1234567890.123456),
}


class TestCondition(unittest.TestCase):
    def test_int_int(self):
        condition, _ = Condition.compile_and_validate(""" 1 == 1 """)
        self.assertTrue(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(""" 1 == 2 """)
        self.assertFalse(condition.evaluate(TestActivation))

    def test_int_other(self):
        condition, _ = Condition.compile_and_validate(""" 1 == "1" """)
        self.assertFalse(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(""" 1 == "2" """)
        self.assertFalse(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(""" 1 == 1.0 """)
        self.assertFalse(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(""" 1 == 2.0 """)
        self.assertFalse(condition.evaluate(TestActivation))

    def test_msg_int(self):
        condition, _ = Condition.compile_and_validate(""" msg.message.code == 200 """)
        self.assertTrue(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(""" msg.message.code == 201 """)
        self.assertFalse(condition.evaluate(TestActivation))

    def test_msg_other(self):
        condition, _ = Condition.compile_and_validate(""" msg.message.code == "200" """)
        self.assertFalse(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(""" msg.message.code == "201" """)
        self.assertFalse(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(""" msg.message.code == 200.0 """)
        self.assertFalse(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(""" msg.message.code == 201.0 """)
        self.assertFalse(condition.evaluate(TestActivation))

    def test_msg_cast_other(self):
        condition, _ = Condition.compile_and_validate(
            """ msg.message.code == int("200") """
        )
        self.assertTrue(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.message.code == int("201") """
        )
        self.assertFalse(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.message.code == int(200.0) """
        )
        self.assertTrue(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.message.code == int(201.0) """
        )
        self.assertFalse(condition.evaluate(TestActivation))

    def test_msg_scope(self):
        condition, _ = Condition.compile_and_validate(
            """ msg.message.code == scope.code """
        )
        self.assertTrue(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.message.code == scope.invalid_attr """
        )
        self.assertFalse(condition.evaluate(TestActivation))

    def test_invalid(self):
        condition, _ = Condition.compile_and_validate(""" mmm.message.code == 200 """)
        self.assertFalse(condition.evaluate(TestActivation))

    def test_map_contains(self):
        condition, _ = Condition.compile_and_validate(
            """ msg.lst.map(x, x.code).contains(2) """
        )
        self.assertTrue(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.lst.map(x, x.code).contains(4) """
        )
        self.assertFalse(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.lst.map(x, x.code * 2).contains(6) """
        )
        self.assertTrue(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.lst.map(x, x.code * 2).contains(7) """
        )
        self.assertFalse(condition.evaluate(TestActivation))

    def test_exists(self):
        condition, _ = Condition.compile_and_validate(
            """ msg.lst.exists(x, x.code == 2) """
        )
        self.assertTrue(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.lst.exists(x, x.code == 4) """
        )
        self.assertFalse(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.lst.exists(x, x.code * 2 == 6) """
        )
        self.assertTrue(condition.evaluate(TestActivation))

        condition, _ = Condition.compile_and_validate(
            """ msg.lst.exists(x, x.code * 2 == 7) """
        )
        self.assertFalse(condition.evaluate(TestActivation))

    def test_compile_and_validate(self):
        _, err = Condition.compile_and_validate("""msg.message.code == """)
        self.assertIsNotNone(err)

        _, err = Condition.compile_and_validate("""msg[0 """)
        self.assertIsNotNone(err)



import unittest
from functools import partial

import celpy

from rule_engine.action import Action, compile_embedded_expr

TestActivation = {
    "msg": celpy.adapter.json_to_cel({"message": {"code": "200"}}),
    "scope": celpy.adapter.json_to_cel({"code": "200"}),
    "topic": celpy.celtypes.StringType("/TestTopic"),
    "ts": celpy.celtypes.DoubleType(1234567890.123456),
}


class TestAction(unittest.TestCase):
    @staticmethod
    def serialize_impl(
        str_arg, int_arg, bool_arg, other_str_arg, dict_arg, action_result: list[str]
    ):
        result = {
            "str_arg": str_arg,
            "int_arg": int_arg,
            "bool_arg": bool_arg,
            "other_str_arg": other_str_arg,
            "dict_arg": dict_arg,
        }
        action_result[0] = str(result)

    @staticmethod
    def get_action(raw_kwargs, action_result):
        return Action.compile_and_validate(
            name="serialize",
            raw_kwargs=raw_kwargs,
            impl=partial(TestAction.serialize_impl, action_result=action_result),
        )

    def test_simple(self):
        action_result = [""]
        action, _ = self.get_action(
            {
                "str_arg": "hello",
                "int_arg": 123,
                "bool_arg": True,
                "other_str_arg": "world",
                "dict_arg": {},
            },
            action_result,
        )

        self.assertEqual("", action_result[0])
        action.run(TestActivation)
        self.assertEqual(
            "{'str_arg': 'hello', 'int_arg': 123, 'bool_arg': True, 'other_str_arg': 'world', 'dict_arg': {}}",
            action_result[0],
        )

    def test_single(self):
        action_result = [""]
        action, _ = self.get_action(
            {
                "str_arg": "aaa{ msg.message.code } bbb",
                "int_arg": 123,
                "bool_arg": True,
                "other_str_arg": "worl{ scope.invalid_attr }d",
                "dict_arg": {},
            },
            action_result,
        )

        self.assertEqual("", action_result[0])
        action.run(TestActivation)
        self.assertEqual(
            "{'str_arg': 'aaa200 bbb', 'int_arg': 123, "
            "'bool_arg': True, 'other_str_arg': 'worl{ ERROR }d', 'dict_arg': {}}",
            action_result[0],
        )

    def test_dict(self):
        action_result = [""]
        action, _ = self.get_action(
            {
                "str_arg": "ccc",
                "int_arg": 123,
                "bool_arg": True,
                "other_str_arg": "worl{ scope.invalid_attr }d",
                "dict_arg": {"aaa": "aaa{ msg.message.code } bbb", "mmm": 1},
            },
            action_result,
        )

        self.assertEqual("", action_result[0])
        action.run(TestActivation)
        self.assertEqual(
            "{'str_arg': 'ccc', 'int_arg': 123, 'bool_arg': True, "
            "'other_str_arg': 'worl{ ERROR }d', 'dict_arg': {'aaa': 'aaa200 bbb', 'mmm': 1}}",
            action_result[0],
        )

    def test_validation(self):
        action, err = self.get_action(
            {
                "str_arg": "hello",
                "int_arg": 123,
                "bool_arg": True,
                "other_str_arg": "world",
            },
            [""],
        )
        self.assertIsNone(err)

        action, err = self.get_action(
            {
                "str_arg": "hello",
                "int_arg": 123,
                "bool_arg": True,
                "other_str_arg": "wor{ 1+ }ld",
            },
            [""],
        )
        self.assertIsNotNone(err)

        action, err = self.get_action(
            {
                "str_arg": "ccc",
                "int_arg": 123,
                "bool_arg": True,
                "other_str_arg": "worl{ scope.invalid_attr }d",
                "dict_arg": {"aaa": {"bbb": 1}},
            },
            [""],
        )
        self.assertIsNotNone(err)


class TestEvaluateEmbeddedExpression(unittest.TestCase):
    def assert_expression(self, expr, activation, expected):
        """
        Helper function
        Assert that the expression evaluates to the expected value
        """
        evaluated = compile_embedded_expr(expr)
        self.assertEqual(evaluated(activation), expected)

    def test_simple(self):
        expression = "A simple expression"
        self.assert_expression(expression, TestActivation, "A simple expression")

    def test_single(self):
        expression = "aaa { msg.code  } bbb"
        self.assert_expression(expression, TestActivation, "aaa { ERROR } bbb")

        expression = "aaa { msg.message.code } bbb"
        self.assert_expression(expression, TestActivation, "aaa 200 bbb")

    def test_multiple(self):
        expression = """
        aaa { msg.code }
        bbb { msg.message.code }
        ccc { scope.code }
        ddd { topic }
        eee { ts }
        """
        expected = """
        aaa { ERROR }
        bbb 200
        ccc 200
        ddd /TestTopic
        eee 1234567890.123456
        """
        self.assert_expression(expression, TestActivation, expected)



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator as op
from abc import ABC, abstractmethod
from functools import wraps


class Condition(ABC):
    """
    Defines a condition to be evaluated on a single message.

    It is assumed that the message has three fields, mirroring those of the ROS
    bag reader API:

    - topic
    - msg
    - msgtype
    - ts

    To make the DSL look nice at the end, this class ends up taking on a lot of
    the complexity. If you're familiar with monads, this class is a combination
    of Future and State and Optional.

    If you're not familiar with monads, this is going to be very confusing. I'm
    sorry.

    """

    @abstractmethod
    def evaluate_condition_at(self, item, scope):
        pass

    @staticmethod
    def wrap(value):
        if isinstance(value, Condition):
            return value
        return ThunkCondition(lambda item, scope: (value, scope))

    @staticmethod
    def wrap_args(func):
        """Decorator that calls wrap on all the args of a function."""

        @wraps(func)
        def result_func(*args, **kwargs):
            args = [Condition.wrap(value) for value in args]
            kwargs = {key: Condition.wrap(value) for key, value in kwargs.items()}
            return func(*args, **kwargs)

        return result_func

    @staticmethod
    def flatmap(self, mapper):
        def new_thunk(item, scope):
            value1, scope = self.evaluate_condition_at(item, scope)
            if value1 is None:
                return value1, scope

            return mapper(value1).evaluate_condition_at(item, scope)

        return ThunkCondition(new_thunk)

    @staticmethod
    def map(self, mapper):
        return MappedCondition(self, mapper)

    @staticmethod
    def apply(func, *args):
        def new_thunk(item, scope):
            new_args = []
            for arg in args:
                v, scope = arg.evaluate_condition_at(item, scope)
                if v is None:
                    return None, scope
                new_args.append(v)
            return func(scope, *new_args)

        return ThunkCondition(new_thunk)

    def __eq__(self, other):
        return self.__wrap_binary_op(other, op.eq)

    def __ne__(self, other):
        return self.__wrap_binary_op(other, op.ne)

    def __gt__(self, other):
        return self.__wrap_binary_op(other, op.gt, float)

    def __ge__(self, other):
        return self.__wrap_binary_op(other, op.ge, float)

    def __lt__(self, other):
        return self.__wrap_binary_op(other, op.lt, float)

    def __le__(self, other):
        return self.__wrap_binary_op(other, op.le, float)

    def __neg__(self):
        return Condition.map(self, op.neg)

    def __pos__(self):
        return Condition.map(self, op.pos)

    def __add__(self, other):
        return self.__wrap_binary_op(other, op.add, float)

    def __radd__(self, other):
        return self.__wrap_binary_op(other, op.add, float, True)

    def __sub__(self, other):
        return self.__wrap_binary_op(other, op.sub, float)

    def __rsub__(self, other):
        return self.__wrap_binary_op(other, op.sub, float, True)

    def __mul__(self, other):
        return self.__wrap_binary_op(other, op.mul, float)

    def __rmul__(self, other):
        return self.__wrap_binary_op(other, op.mul, float, True)

    def __truediv__(self, other):
        return self.__wrap_binary_op(other, op.truediv, float)

    def __rtruediv__(self, other):
        return self.__wrap_binary_op(other, op.truediv, float, True)

    def __call__(self, *args, **kwargs):
        return Condition.map(self, lambda f: f(*args, **kwargs))

    def __getattr__(self, name):
        """Priority: __getattr__ > __getitem__"""
        return Condition.map(self, get_attr_or_item(name))

    def __getitem__(self, item):
        """Priority: __getitem__ > __getattr__"""
        return Condition.map(self, get_item_or_attr(item))

    def __wrap_binary_op(self, other, op, coerce=lambda x: x, swap=False):
        other = Condition.wrap(other)
        return Condition.flatmap(
            self,
            lambda x: Condition.map(
                other,
                lambda y: (
                    op(coerce(y), coerce(x)) if swap else op(coerce(x), coerce(y))
                ),
            ),
        )

    def __bool__(self):
        raise Exception(
            """
            It is intentional that Condition objects should not be used as boolean values.

            If you're trying to use boolean operators with conditions, please use function equivalents instead:
              a and b -> and_(a, b)
              a or b -> or_(a, b)
              not a -> not_(a)

            If you're trying to use the `in` operator, please also use function equivalent instead:
              a in b -> has(b, a)
            """
        )


def get_attr_or_item(name):
    def __get_attr_or_item(x):
        try:
            if hasattr(x, name):
                return getattr(x, name)
            if hasattr(x, "__getitem__"):
                return x[name]
        except (KeyError, IndexError, TypeError):
            return None
        return None

    return __get_attr_or_item


def get_item_or_attr(item):
    def __get_item_or_attr(x):
        try:
            if hasattr(x, "__getitem__"):
                return x[item]
            if hasattr(x, item):
                return getattr(x, item)
        except (KeyError, IndexError, TypeError):
            return None
        return None

    return __get_item_or_attr


class ThunkCondition(Condition):
    def __init__(self, thunk):
        super().__init__()
        self.__thunk = thunk

    def evaluate_condition_at(self, item, scope):
        return self.__thunk(item, scope)


class MappedCondition(Condition):
    def __init__(self, inner, mapper):
        super().__init__()
        self.__inner = inner
        self.__mapper = mapper

    def evaluate_condition_at(self, item, scope):
        value, scope = self.__inner.evaluate_condition_at(item, scope)
        if value is not None:
            value = self.__mapper(value)

        return value, scope


__all__ = [
    "Condition",
    "ThunkCondition",
]



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import re
from functools import partial
from typing import Callable

import celpy

from rule_engine.utils import ENV, log_level_decorator


class Action:
    """
    Defining an action, of which the underlying implementation is a function
    """

    @staticmethod
    @log_level_decorator(logging.WARN)
    def compile_and_validate(name: str, raw_kwargs: dict[str, any], impl: Callable):
        """
        Compile and validate the action
        """
        try:
            kwargs = {k: compile_value(v) for k, v in raw_kwargs.items()}
            return Action(name, raw_kwargs, kwargs, impl), None
        except Exception as e:
            return None, e

    def __init__(
        self,
        name: str,
        raw_kwargs: dict[str, any],
        kwargs: dict[str, Callable[[celpy.Context], any]],
        impl: Callable,
    ):
        self.name = name
        self.raw_kwargs = raw_kwargs
        self._impl = impl
        self._kwargs = kwargs

    @log_level_decorator(logging.WARN)
    def run(self, activation: celpy.Context):
        """
        Run the action with the activation dictionary
        """
        self._impl(**{k: v(activation) for k, v in self._kwargs.items()})

    def __repr__(self):
        return f"Action({self.name}){self.raw_kwargs}"


def compile_value(value: any) -> Callable[[celpy.Context], any]:
    """
    Compile a value, which can be a string that optionally with embedded CEL expression,
    or otherwise a constant value
    """
    if isinstance(value, str):
        return compile_embedded_expr(value)
    elif isinstance(value, dict):
        compiled_value = {}
        for k, v in value.items():
            if isinstance(v, dict):
                raise ValueError("Nested dict is not supported")
            compiled_value[k] = compile_value(v)

        def evaluate_dict(activation: celpy.Context, _value: dict):
            return {_k: _v(activation) for _k, _v in _value.items()}

        return partial(evaluate_dict, _value=compiled_value)

    else:

        def wrap(_value):
            return lambda _: _value

        return wrap(value)


def compile_embedded_expr(expr: str) -> Callable[[celpy.Context], str]:
    """
    Compile a string with optional embedded CEL expression, for example:
    expression: "aaa { msg.code } bbb { msg.error } {msg.aaa}"
    activation: {"msg": {"code": "404", "error": "Not found"}}
    result: "aaa 404 bbb Not found { ERROR }"

    The evaluation process is done by evaluating all the CEL expression enclosed in {{ and }} and
    replacing the expression with the evaluated value. (Trim pre- and post- spaces in between {{ and }})

    Returns a function that takes an activation dictionary and returns the evaluated string
    """
    pattern = re.compile(r"\{\s*(.*?)\s*}")
    matches = pattern.findall(expr)
    compiled_programs = [ENV.program(ENV.compile(match)) for match in matches]

    def evaluate(
        activation: celpy.Context, expression: str, programs: list[celpy.Runner]
    ) -> str:
        # Replace the matches in the expression with the evaluated value
        idx = -1

        def replace(_match):
            nonlocal idx
            idx += 1
            try:
                return str(programs[idx].evaluate(activation))
            except Exception:
                return "{ ERROR }"

        return re.sub(pattern, replace, expression)

    return partial(evaluate, expression=expr, programs=compiled_programs)



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from dataclasses import dataclass, field
from typing import Any

_log = logging.getLogger(__name__)


@dataclass
class DiagnosisItem:
    topic: str
    msg: Any
    ts: float
    msgtype: str


@dataclass
class Rule:
    conditions: list
    actions: list
    initial_scope: dict
    upload_limit: dict = field(default_factory=dict)
    spec: dict = field(default_factory=dict)
    project_name: str = ""


class Engine:
    def __init__(self, rules, should_trigger_action=None, trigger_cb=None):
        self.__rules = rules
        self.__should_trigger_action = should_trigger_action
        self.__trigger_cb = trigger_cb

    def consume_next(self, item):
        for rule in self.__rules:
            triggered_condition_indices = []
            triggered_scope = None

            for i, cond in enumerate(rule.conditions):
                res, scope = cond.evaluate_condition_at(item, rule.initial_scope)
                _log.debug(f"evaluate condition, result: {res}, scope: {scope}")
                if res:
                    triggered_condition_indices.append(i)
                if not triggered_scope:
                    triggered_scope = scope

            if not triggered_condition_indices:
                continue

            # For testing, rule.spec is not specified
            hit = (
                {}
                if not rule.spec
                else {
                    **rule.spec,
                    "when": [rule.spec["when"][i] for i in triggered_condition_indices],
                }
            )

            action_triggered = False
            if not self.__should_trigger_action or self.__should_trigger_action(
                rule.project_name, rule.spec, hit
            ):
                action_triggered = True
                for action in rule.actions:
                    action.run(item, triggered_scope)

            if self.__trigger_cb:
                self.__trigger_cb(
                    rule.project_name, rule.spec, hit, action_triggered, item
                )



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast

from ruleengine.dsl.validation.normalizer import normalize_expression_tree
from ruleengine.dsl.validation.validator import base_dsl_values


def str_to_condition(expr_str, additional_injected_values=None):
    parsed = ast.parse(expr_str, mode="eval")
    code = compile(normalize_expression_tree(parsed), "", mode="eval")
    return eval(code, {**base_dsl_values, **(additional_injected_values or {})})



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast

from .normalizer import normalize_expression_tree
from .validation_result import ValidationErrorType, ValidationResult


def validate_expression(expr_str, injected_values):
    try:
        parsed = ast.parse(expr_str, mode="eval")
    except SyntaxError:
        return ValidationResult(False, ValidationErrorType.SYNTAX)

    for node in ast.walk(parsed):
        if isinstance(node, ast.Name):
            name = node.id
            if name not in injected_values:
                return ValidationResult(
                    False, ValidationErrorType.UNDEFINED, {"name": name}
                )

    code = compile(normalize_expression_tree(parsed), "", mode="eval")

    return ValidationResult(True, entity=eval(code, injected_values))



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import celpy

from rule_engine.utils import ENV, log_level_decorator


class Condition:
    """
    Defining a condition with a CEL expression
    """

    @staticmethod
    @log_level_decorator(logging.WARN)
    def compile_and_validate(raw_condition: str):
        """
        Validate the condition
        """
        try:
            program = ENV.program(ENV.compile(raw_condition))
            return Condition(raw_condition, program), None
        except Exception as e:
            return None, e

    def __init__(self, raw: str, program: celpy.Runner):
        self.raw = raw
        self.program = program

    @log_level_decorator(logging.WARN)
    def evaluate(self, activation: celpy.Context):
        """
        Evaluate the condition as boolean
        """
        try:
            # Restrict the return to be true when the condition evaluates exactly to boolean true
            return self.program.evaluate(activation).__repr__() == "BoolType(True)"
        except Exception:
            return False

    def __repr__(self):
        return f"Condition({self.raw})"

    def __str__(self):
        return f"Condition({self.raw})"



# Copyright 2024 coScene
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from ruleengine.dsl import base_conditions, log_conditions, sequence_conditions
from ruleengine.dsl.action import Action
from ruleengine.dsl.condition import Condition
from .actions import ActionValidator, UnknownFunctionKeywordArgException
from .ast import validate_expression
from .validation_result import ValidationErrorType, ValidationResult

base_dsl_values = dict(
    inspect.getmembers(base_conditions)
    + inspect.getmembers(log_conditions)
    + inspect.getmembers(sequence_conditions)
)


def validate_condition(cond_str):
    return _do_validate(
        cond_str, base_dsl_values, Condition, ValidationErrorType.NOT_CONDITION
    )


def validate_action(action_str, action_impls):
    action_validator = ActionValidator(action_impls)
    action_dsl_values = {
        "upload": action_validator.upload,
        "create_moment": action_validator.create_moment,
        **base_dsl_values,
    }
    return _do_validate(
        action_str,
        action_dsl_values,
        Action,
        ValidationErrorType.NOT_ACTION,
    )


def _do_validate(expr_str, injected_values, expected_class, class_expectation_error):
    if not expr_str.strip():
        return ValidationResult(False, ValidationErrorType.EMPTY)

    try:
        res = validate_expression(expr_str, injected_values)
        if not res.success:
            return res
    except UnknownFunctionKeywordArgException as e:
        return ValidationResult(False, ValidationErrorType.UNDEFINED, {"name": e.name})
    except Exception as e:
        return ValidationResult(False, ValidationErrorType.UNKNOWN, {"message": str(e)})

    if not isinstance(res.entity, expected_class):
        return ValidationResult(
            False, class_expectation_error, {"actual": type(res.entity).__name__}
        )

    return res
